# third party imports
import numpy as np
import pandas as pd
# local imports
from datasurveyor import GeneralFeatures as gf


# should pass all checks
good = pd.DataFrame.from_dict({
    'g1': (True, False, True, False),
    'g2': (0, 1, 0, 1),
    'g3': (False, True, False, True),
})

# should fail null check
bad_nulls1 = pd.DataFrame.from_dict({
    'b1': (True, np.NaN, False, True),
    'b2': (1, 0, 1, 0),
})

# should fail fuzzy null check
bad_fuzzy1 = pd.DataFrame.from_dict({
    'b1': ('True', '', 'False', 'True'),
    'b2': (1, 0, 1, 0),
})

# should fail fuzzy null check
bad_fuzzy2 = pd.DataFrame.from_dict({
    'b1': ('True', 'NULL', 'False', 'True'),
    'b2': (1, 0, 1, 0),
})

# should fail fuzzy null check with added fuzzy
bad_fuzzy3 = pd.DataFrame.from_dict({
    'b1': ('True', 'foo', 'False', 'True'),
    'b2': (1, 0, 1, 0),
})


# fail DataFrames for fuzzy nulls
fail_fuzzy_df = pd.DataFrame.from_dict({
    'column': ('b1', 'b2'),
    'fuzzy_nulls_present': (True, False),
    'fuzzy_null_count': (1, 0),
    'prop_fuzzy_null': (0.25, 0.0)
})
fail_fuzzy_ser = pd.DataFrame.from_dict({
    'fuzzy_nulls_present': (True,),
    'fuzzy_null_count': (1,),
    'prop_fuzzy_null': (0.25,),
})


def test_check_nulls_good_df():
    # verifies good data passes the null check
    cols = ['nulls_present', 'null_count', 'prop_null']
    assert not gf.check_nulls(good).loc[:, cols].any(axis=None)


def test_check_nulls_good_ser():
    # verifies good data passes the null check
    cols = ['null_count', 'prop_null']
    assert not gf.check_nulls(good['g1']).loc[:, cols].any(axis=None)


def test_check_nulls_bad_df():
    # verifies that the null check finds rows with nulls
    fail = pd.DataFrame.from_dict({
        'column': ('b1', 'b2'),
        'nulls_present': (True, False),
        'null_count': (1, 0),
        'prop_null': (0.25, 0.0)
    })
    assert fail.equals(gf.check_nulls(bad_nulls1))


def test_check_nulls_bad_ser():
    # verifies that the null check finds rows with nulls
    fail = pd.DataFrame.from_dict({
        'nulls_present': (True,),
        'null_count': (1,),
        'prop_null': (0.25,),
    })
    assert fail.equals(gf.check_nulls(bad_nulls1['b1']))


def test_fuzzy_nulls_good_df():
    # verifies good data passes the null check
    cols = ['fuzzy_nulls_present', 'fuzzy_null_count', 'prop_fuzzy_null']
    assert not gf.check_fuzzy_nulls(good).loc[:, cols].any(axis=None)


def test_fuzzy_nulls_good_ser():
    # verifies good data passes the null check
    cols = ['fuzzy_nulls_present', 'fuzzy_null_count', 'prop_fuzzy_null']
    assert not gf.check_fuzzy_nulls(good['g1']).loc[:, cols].any(axis=None)


def test_fuzzy_nulls_bad1_df():
    # verifies that the fuzzy null check finds rows with fuzzy nulls
    assert fail_fuzzy_df.equals(gf.check_fuzzy_nulls(bad_fuzzy1))


def test_fuzzy_nulls_bad1_ser():
    # verifies that the fuzzy null check finds rows with fuzzy nulls
    assert fail_fuzzy_ser.equals(gf.check_fuzzy_nulls(bad_fuzzy1['b1']))


def test_fuzzy_nulls_bad2_df():
    # verifies that the fuzzy null check finds rows with fuzzy nulls
    assert fail_fuzzy_df.equals(gf.check_fuzzy_nulls(bad_fuzzy2))


def test_fuzzy_nulls_bad2_ser():
    # verifies that the fuzzy null check finds rows with fuzzy nulls
    assert fail_fuzzy_ser.equals(gf.check_fuzzy_nulls(bad_fuzzy2['b1']))


def test_fuzzy_nulls_bad3_df():
    # verifies that the fuzzy null check finds rows with added fuzzy nulls
    assert fail_fuzzy_df.equals(gf.check_fuzzy_nulls(bad_fuzzy3, add_fuzzy_nulls=['foo']))


def test_fuzzy_nulls_bad3_ser():
    # verifies that the fuzzy null check finds rows with added fuzzy nulls
    assert fail_fuzzy_ser.equals(gf.check_fuzzy_nulls(bad_fuzzy3['b1'], add_fuzzy_nulls=['foo']))



"""Data exploration tools."""

__version__ = '0.0.1'


# Binary feature checks
from datasurveyor._binary_features import BinaryFeatures
# Categorical feature checks
from datasurveyor._categorical_features import CategoricalFeatures
# General feature checks
from datasurveyor._general_features import GeneralFeatures
# Unique feature checks
from datasurveyor._unique_features import UniqueFeatures


__all__ = [_binary_features, _categorical_features, _general_features, _unique_features]



# third party imports
import pandas as pd
import pytest
# local imports
from datasurveyor import BinaryFeatures as bf


# should pass all checks
good = pd.DataFrame.from_dict({
    'g1': (True, False, True, False),
    'g2': (0, 1, 0, 1),
    'g3': (False, True, False, True),
})

# should fail dtype check due to strings
bad_type1 = pd.DataFrame.from_dict({
    'b1': ('True', 'True', 'False', 'True'),
    'b2': (False, True, False, True),
})

# should fail dtype check due to floats
bad_type2 = pd.DataFrame.from_dict({
    'b1': (1.0, 0.0, 0.6, 0.9),
    'b2': (False, True, False, True),
})

# should fail all same check
bad_same1 = pd.DataFrame.from_dict({
    'b1': (1, 1, 1, 1),
    'b2': (False, True, False, True),
})

# should fail mostly same check with threshold 0.7
bad_same2 = pd.DataFrame.from_dict({
    'b1': (True, True, True, False),
    'b2': (False, True, False, True),
})

# should fail range check (less than 0)
bad_range1 = pd.DataFrame.from_dict({
    'b1': (1, 0, 0, -1),
    'b2': (False, True, False, True),
})

# should fail range check (greater than 1)
bad_range2 = pd.DataFrame.from_dict({
    'b1': (1, 0, 10, 1),
    'b2': (False, True, False, True),
})


def test_validate_dtype_good_df():
    # verifies good data passes the dtype check
    bf._validate_binary_dtype(good)


def test_validate_dtype_good_ser():
    # verifies good data passes the dtype check
    bf._validate_binary_dtype(good['g1'])


def test_validate_dtype_bad_str_df():
    # checks that TypeError is raised when df contains object (str) data
    with pytest.raises(TypeError) as excinfo:
        bf._validate_binary_dtype(bad_type1)
    # verifies TypeError contains appropriate message
    assert 'should be of type bool or int64' in str(excinfo.value)


def test_validate_dtype_bad_str_ser():
    # checks that TypeError is raised when series contains object (str) data
    with pytest.raises(TypeError) as excinfo:
        bf._validate_binary_dtype(bad_type1['b1'])
    # verifies TypeError contains appropriate message
    assert 'should be of type bool or int64' in str(excinfo.value)


def test_validate_dtype_bad_float_df():
    # checks that TypeError is raised when df contains float data
    with pytest.raises(TypeError) as excinfo:
        bf._validate_binary_dtype(bad_type2)
    # verifies TypeError contains appropriate message
    assert 'should be of type bool or int64' in str(excinfo.value)


def test_validate_dtype_bad_float_ser():
    # checks that TypeError is raised when series contains float data
    with pytest.raises(TypeError) as excinfo:
        bf._validate_binary_dtype(bad_type2['b1'])
    # verifies TypeError contains appropriate message
    assert 'should be of type bool or int64' in str(excinfo.value)


def test_check_all_same_good_df():
    # verifies good data passes the all same check
    assert not bf.check_all_same(good).loc[:, 'all_same'].any()


def test_check_all_same_good_ser():
    # verifies good data passes the all same check
    assert not bf.check_all_same(good['g1']).loc[:, 'all_same'].any()


def test_check_all_same_bad_df():
    # verifies that the all same check finds rows with all the same value
    fail = pd.DataFrame.from_dict({'column': ('b1', 'b2'), 'all_same': (True, False)})
    assert fail.equals(bf.check_all_same(bad_same1))


def test_check_all_same_bad_ser():
    # verifies that the all same check finds rows with all the same value
    fail = pd.DataFrame.from_dict({'all_same': (True,)})
    assert fail.equals(bf.check_all_same(bad_same1['b1']))


def test_check_mostly_same_good_df():
    # verifies good data passes the mostly same check
    assert not bf.check_mostly_same(good).loc[:, 'mostly_same'].any(axis=0)


def test_check_mostly_same_good_ser():
    # verifies good data passes the mostly same check
    assert not bf.check_mostly_same(good['g1']).loc[:, 'mostly_same'].any()


def test_check_mostly_same_bad1_df():
    # verifies that the mostly same check finds rows with mostly the same value
    fail = pd.DataFrame.from_dict({
        'column': ('b1', 'b2'),
        'mostly_same': (True, False),
        'thresh': (0.7, 0.7),
        'mean': (0.75, 0.5),
    })
    assert fail.equals(bf.check_mostly_same(bad_same2, 0.7))


def test_check_mostly_same_bad1_ser():
    # verifies that the mostly same check finds rows with mostly the same value
    fail = pd.DataFrame.from_dict({
        'mostly_same': (True,),
        'thresh': (0.7,),
        'mean': (0.75,),
    })
    assert fail.equals(bf.check_mostly_same(bad_same2['b1'], 0.7))


def test_check_outside_range_good_df():
    # verifies good data passes the range check
    assert not bf.check_outside_range(good).loc[:, 'outside_range'].any(axis=0)


def test_check_range_good_ser():
    # verifies good data passes the range check
    assert not bf.check_outside_range(good['g1']).loc[:, 'outside_range'].any()


def test_check_range_bad1_df():
    # verifies that the range check finds rows with low values
    fail = pd.DataFrame.from_dict({'column': ('b1', 'b2'), 'outside_range': (True, False)})
    assert fail.equals(bf.check_outside_range(bad_range1))


def test_check_range_bad1_ser():
    # verifies that the range check finds rows with low values
    fail = pd.DataFrame.from_dict({'outside_range': (True,)})
    assert fail.equals(bf.check_outside_range(bad_range1['b1']))


def test_check_range_bad2_df():
    # verifies that the range check finds rows with high values
    fail = pd.DataFrame.from_dict({'column': ('b1', 'b2'), 'outside_range': (True, False)})
    assert fail.equals(bf.check_outside_range(bad_range2))


def test_check_range_bad2_ser():
    # verifies that the range check finds rows with high values
    fail = pd.DataFrame.from_dict({'outside_range': (True,)})
    assert fail.equals(bf.check_outside_range(bad_range2['b1']))



# standard library imports
from typing import Union
# third party imports
import numpy as np
import pandas as pd
# local imports
from datasurveyor import _utils


class UniqueFeatures:

    @staticmethod
    def _validate_unique_dtype(data: Union[pd.DataFrame, pd.Series]) -> None:
        """Validates that unique data contains only dtype object, int, or datetime.

        Args:
            data: Unique data to be validated.

        Returns:
            None

        Raises:
            TypeError: If `data` contains dtype other than object, int, or datetime.
        """
        is_df = _utils.check_if_df(data)
        err_message = 'Unique feature columns should be of type object, int64, or datetime64.'
        types = (np.dtype('O'), np.dtype(int), np.dtype('datetime64[ns]'))
        if is_df:
            if not data.dtypes.isin(types).all():
                raise TypeError(err_message)
        else:
            if data.dtypes not in types:
                raise TypeError(err_message)
        return

    @staticmethod
    def check_uniqueness(
            data: Union[pd.DataFrame, pd.Series],
    ) -> pd.DataFrame:
        """Checks if unique data contains columns with duplicates.

        Args:
            data: Data to be checked for duplicates.

        Returns:
            DataFrame with bool(s) indicating if data contains duplicates, the count of
            duplicates present, and the proportion of duplicates.

        Raises:
            ValueError: If unique data contains nulls.
        """
        UniqueFeatures._validate_unique_dtype(data)
        is_df = _utils.check_if_df(data)
        err_message = 'Columns with unique data should not contain nulls.'
        if is_df:
            if data.isna().any(axis=None):
                raise ValueError(err_message)
            count_dupes = data.nunique(axis=0).subtract(data.shape[0]).multiply(-1)
            is_dupes = count_dupes.astype(bool)
            prop_dupes = count_dupes.divide(data.shape[0])
        else:
            if data.isna().any():
                raise ValueError(err_message)
            count_dupes = data.shape[0] - data.nunique()
            is_dupes = bool(count_dupes)
            prop_dupes = count_dupes / data.shape[0]
        result = _utils.result_to_df(
            data=is_dupes,
            title='dupes_present',
            dupe_count=count_dupes,
            prop_dupe=prop_dupes
        )
        return result



# standard library imports
from typing import Union
# third party imports
import numpy as np
import pandas as pd
# local imports
from datasurveyor import _utils


class CategoricalFeatures:

    @staticmethod
    def _validate_categorical_dtype(data: Union[pd.DataFrame, pd.Series]) -> None:
        """Validates that categorical data contains only dtype int or object (str).

        Args:
            data: Categorical data to be type validated.

        Returns:
            None

        Raises:
            TypeError: If `data` contains dtype other than bool or object (str).
        """
        is_df = _utils.check_if_df(data)
        err_message = 'Unique feature columns should be of type object or int64.'
        types = (np.dtype('O'), np.dtype(int))
        if is_df:
            if not data.dtypes.isin(types).all():
                raise TypeError(err_message)
        else:
            if data.dtypes not in types:
                raise TypeError(err_message)
        return

    @staticmethod
    def check_mostly_same(
            data: Union[pd.DataFrame, pd.Series],
            thresh: float = 0.95,
            dropna: bool = False,
    ) -> pd.DataFrame:
        """Checks if categorical data contains almost all the same category.

        Args:
            data: Categorical data to be checked if almost all the same category.
            thresh: Threshold for what proportion of data must be the same category to fail check.
            dropna: If True: ignores nulls, if False: counts nulls as a category.

        Returns:
            DataFrame with bool(s) indicating if data contains almost all the same category, the
            value of threshold used to determine if mostly same, the most common category, the
            count of the most common category, and the proportion of the most common category.
        """
        _utils.validate_thresh(thresh)
        CategoricalFeatures._validate_categorical_dtype(data)
        is_df = _utils.check_if_df(data)
        if is_df:
            most_common = data.mode(axis=0, dropna=dropna).loc[0, :]
            count_common = data.eq(most_common).sum(axis=0)
            prop_common = count_common.divide(data.shape[0])
            mostly_same = prop_common.ge(thresh)
        else:
            most_common = data.mode()[0]
            count_common = data.eq(most_common).sum()
            prop_common = count_common / data.shape[0]
            mostly_same = prop_common >= thresh
        result = _utils.result_to_df(
            mostly_same,
            title='mostly_same',
            thresh=thresh,
            most_common=most_common,
            count=count_common,
            prop=prop_common,
        )
        return result

    @staticmethod
    def check_n_categories(
            data: Union[pd.DataFrame, pd.Series],
            dropna: bool = False,
    ) -> pd.DataFrame:
        """Counts the number of categories.

        Args:
            data: Data to count categories for.
            dropna: If True: ignores nulls, if False: counts nulls as a category.

        Returns:
            DataFrame with count(s) of categories.
        """
        CategoricalFeatures._validate_categorical_dtype(data)
        is_df = _utils.check_if_df(data)
        if is_df:
            result = data.nunique(axis=0, dropna=dropna)
        else:
            result = data.nunique(dropna=dropna)
        return _utils.result_to_df(result, title='n_categories')



# third party imports
import numpy as np
import pandas as pd
import pytest
# local imports
from datasurveyor import CategoricalFeatures as cf


# should pass all checks
good = pd.DataFrame.from_dict({
    'g1': ('a', 'b', 'c', 'd'),
    'g2': (0, 1, 1, 0),
})

# used to test behavior of data with nan's
nan = pd.DataFrame.from_dict({
    'g1': (np.nan, np.nan, 'a', 'b'),
    'g2': (0, 1, 2, 3),
})

# should fail dtype validation
bad_type = pd.DataFrame.from_dict({
    'b1': (True, True, False, True),
    'b2': (1.0, 0.0, 1.0, 0.0),
})

# DataFrames that should match output of check_n_categories
mostly_same_out1 = pd.DataFrame.from_dict({
    'column': ('g1', 'g2'),
    'mostly_same': (False, True),
    'thresh': (0.4, 0.4),
    'most_common': ('a', 0),
    'count': (1, 2),
    'prop': (0.25, 0.5),
})
# DataFrames that should match output of check_n_categories
mostly_same_out2 = pd.DataFrame.from_dict({
    'mostly_same': (True,),
    'thresh': (0.4,),
    'most_common': (0,),
    'count': (2,),
    'prop': (0.5,),
})
good_result = pd.DataFrame.from_dict({
    'column': ('g1', 'g2'),
    'n_categories': (4, 2)
})
drop_nan_result = pd.DataFrame.from_dict({
    'column': ('g1', 'g2'),
    'n_categories': (2, 4)
})
nan_result = pd.DataFrame.from_dict({
    'column': ('g1', 'g2'),
    'n_categories': (3, 4)
})


def test_validate_categorical_dtype_good_df():
    # verifies good data passes the dtype check
    cf._validate_categorical_dtype(good)


def test_validate_categorical_dtype_good_ser():
    # verifies good data passes the dtype check
    cf._validate_categorical_dtype(good['g1'])


def test_validate_categorical_dtype_bad_df():
    # checks that TypeError is raised when df contains float data
    with pytest.raises(TypeError) as excinfo:
        cf._validate_categorical_dtype(bad_type)
    # verifies TypeError contains appropriate message
    assert 'should be of type object or int64' in str(excinfo.value)


def test_validate_categorical_dtype_bad_ser():
    # checks that TypeError is raised when df contains float data
    with pytest.raises(TypeError) as excinfo:
        cf._validate_categorical_dtype(bad_type['b1'])
    # verifies TypeError contains appropriate message
    assert 'should be of type object or int64' in str(excinfo.value)


def test_mostly_same_df():
    # verifies mostly same output matches expectation
    assert mostly_same_out1.equals(cf.check_mostly_same(good, thresh=0.4))


def test_mostly_same_ser():
    # verifies mostly same output matches expectation
    cols = ['mostly_same', 'thresh', 'most_common', 'count', 'prop']
    assert mostly_same_out2.equals(cf.check_mostly_same(good['g2'], thresh=0.4))


def test_check_n_categories_df():
    # verifies that check_n_categories generates the expected output
    assert good_result.equals(cf.check_n_categories(good))


def test_check_n_categories_ser():
    # verifies that check_n_categories generates the expected output
    assert good_result.loc[1, 'n_categories'] == cf.check_n_categories(good['g2']).values


def test_check_n_categories_dropna_df():
    # verifies that check_n_categories generates the expected output with nulls
    assert drop_nan_result.equals(cf.check_n_categories(nan, dropna=True))


def test_check_n_categories_no_dropna_df():
    # verifies that check_n_categories generates the expected output with nulls
    assert nan_result.equals(cf.check_n_categories(nan, dropna=False))


def test_check_n_categories_dropna_ser():
    # verifies that check_n_categories generates the expected output with nulls
    assert drop_nan_result.loc[0, 'n_categories'] == cf.check_n_categories(nan['g1'], dropna=True).values


def test_check_n_categories_no_dropna_ser():
    # verifies that check_n_categories generates the expected output with nulls
    assert nan_result.loc[0, 'n_categories'] == cf.check_n_categories(nan['g1'], dropna=False).values



# standard library imports
from typing import Union
# third party imports
import numpy as np
import pandas as pd
# local imports
from datasurveyor import _utils


class BinaryFeatures:

    @staticmethod
    def _validate_binary_dtype(data: Union[pd.DataFrame, pd.Series]) -> None:
        """Validates that binary data contains only dtype bool or int.

        Args:
            data: Binary data to be type validated.

        Returns:
            None

        Raises:
            TypeError: If `data` contains dtype other than bool or int.
        """
        is_df = _utils.check_if_df(data)
        err_message = 'Binary feature columns should be of type bool or int64.'
        types = (np.dtype(bool), np.dtype(int))
        if is_df:
            if not data.dtypes.isin(types).all():
                raise TypeError(err_message)
        else:
            if data.dtypes not in types:
                raise TypeError(err_message)
        return

    @staticmethod
    def check_all_same(data: Union[pd.DataFrame, pd.Series]) -> Union[pd.DataFrame]:
        """Checks if binary data contains all the same value.

        Args:
            data: Binary data to be checked if all values are the same.

        Returns:
            DataFrame with bool(s) indicating if data contains all the same value.
        """
        is_df = _utils.check_if_df(data)
        BinaryFeatures._validate_binary_dtype(data)
        if is_df:
            result = data.min(axis=0).eq(data.max(axis=0))
        else:
            result = data.min() == data.max()
        return _utils.result_to_df(result, title='all_same')

    @staticmethod
    def check_mostly_same(
            data: Union[pd.DataFrame, pd.Series],
            thresh: float = 0.95,
    ) -> pd.DataFrame:
        """Checks if binary data contains almost all the same value.

        Args:
            data: Binary data to be checked if almost all values are the same.
            thresh: Threshold for what proportion of data must be the same to fail check.

        Returns:
            DataFrame with bool(s) indicating if data contains all the same value, the
            value of threshold used to determine if mostly same, and the average value(s).

        Raises:
            ValueError: If `thresh` less than or equal to 0.0 or greater than or equal to 1.0.
        """
        _utils.validate_thresh(thresh)
        is_df = _utils.check_if_df(data)
        BinaryFeatures._validate_binary_dtype(data)
        if is_df:
            mean = data.mean(axis=0)
            result = (mean >= thresh) | (mean <= 1 - thresh)
        else:
            mean = data.mean()
            result = mean >= thresh or mean <= 1 - thresh
        return _utils.result_to_df(data=result, title='mostly_same', thresh=thresh, mean=mean)

    @staticmethod
    def check_outside_range(data: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
        """Checks if binary data contains columns where min is less than 0 or max is greater than 1.

        Args:
            data: Binary data to be checked if any values are less than 0 or greater than 1.

        Returns:
            DataFrame with bool(s) indicating if data contains any values outside of the expected range.
        """
        is_df = _utils.check_if_df(data)
        BinaryFeatures._validate_binary_dtype(data)
        if is_df:
            result = (data.min(axis=0) < 0) | (data.max(axis=0) > 1)
        else:
            result = data.min() < 0 or data.max() > 1
        return _utils.result_to_df(data=result, title='outside_range')



# standard library imports
from typing import Any, Union
# third party imports
import pandas as pd


def check_if_df(data: Any) -> bool:
    """Checks type of data and raises error if not Pandas DataFrame or Series.

    Args:
        data: Data to be type checked.
    Returns:
        True if `data` is of type Pandas DataFrame.
        False if `data` is of type Pandas Series.

    Raises:
        TypeError if  `data` is not of type pandas DataFrame or Series.
    """
    if isinstance(data, pd.DataFrame):
        return True
    if isinstance(data, pd.Series):
        return False
    else:
        raise TypeError('Input data must be of type pandas DataFrame or pandas Series.')


def result_to_df(
        data: Union[pd.Series, bool, float, int],
        title: str,
        **kwargs,
) -> pd.DataFrame:
    """Generates an output DataFrame with tidy formatting.

    Args:
        data: Base data to be included in DataFrame.
        title: Name of test being run (used as a column name in output DataFrame).
        **kwargs: Additional data to be included in the output DataFrame where:
        argument name = output column name and argument value = output column value.

    Returns:
        Output data as a DataFrame.
    """
    if isinstance(data, pd.Series):
        d = {'column': data.index, title: data.values}
    else:
        d = {title: (data,)}
        if kwargs:
            kwargs = {k: (v,) for k, v in kwargs.items()}
    d.update(kwargs)
    return pd.DataFrame.from_dict(d).reset_index(drop=True)


def validate_thresh(thresh: float) -> None:
    """Validates the proportion provided.

    Args:
        thresh: Proportion to be validated.

    Returns:
        None

    Raises:
        ValueError if `thresh` is not between 0.0 and 1.0 (inclusive).
    """
    if thresh >= 1.0 or thresh <= 0.0:
        raise ValueError('The thresh parameter must be greater than 0.0 and less than 1.0.')
    return



# third party imports
import numpy as np
import pandas as pd
import pytest
# local imports
from datasurveyor import UniqueFeatures as uf


# should pass all checks
good = pd.DataFrame.from_dict({
    'g1': ('foo', 'bar', 'baz', 'boo'),
    'g2': (0, 1, 2, 4),
    'g3': (np.datetime64('2017-01-01'), np.datetime64('2018-01-01'),
           np.datetime64('2019-01-01'), np.datetime64('2020-01-01')),
})

# should fail dtype test
bad_type = pd.DataFrame.from_dict({
    'b1': (0.0, 1.0, 2.0, 4.0),
    'b2': ('foo', 'bar', 'baz', 'boo'),
})

# should fail duplicates test
bad_unique = pd.DataFrame.from_dict({
    'b1': (0, 1, 0, 1),
    'b2': ('foo', 'bar', 'baz', 'boo'),
})

# should raise ValueError
bad_unique_nan = pd.DataFrame.from_dict({
    'b1': ('foo', 'bar', 'baz', np.nan),
    'b2': (0, 1, 2, 3),
})

# fail DataFrames for fuzzy nulls
fail_dupe_df = pd.DataFrame.from_dict({
    'column': ('b1', 'b2'),
    'dupes_present': (True, False),
    'dupe_count': (2, 0),
    'prop_dupe': (0.5, 0.0)
})
fail_dupe_ser = pd.DataFrame.from_dict({
    'dupes_present': (True,),
    'dupe_count': (2,),
    'prop_dupe': (0.5,),
})


def test_validate_dtype_good_df():
    # verifies good data passes the dtype check
    uf._validate_unique_dtype(good)


def test_validate_dtype_good_ser():
    # verifies good data passes the dtype check
    uf._validate_unique_dtype(good['g1'])


def test_validate_dtype_bad_df():
    # checks that TypeError is raised when df contains float data
    with pytest.raises(TypeError) as excinfo:
        uf._validate_unique_dtype(bad_type)
    # verifies TypeError contains appropriate message
    assert 'should be of type object, int64, or datetime64' in str(excinfo.value)


def test_validate_dtype_bad_ser():
    # checks that TypeError is raised when df contains float data
    with pytest.raises(TypeError) as excinfo:
        uf._validate_unique_dtype(bad_type['b1'])
    # verifies TypeError contains appropriate message
    assert 'should be of type object, int64, or datetime64' in str(excinfo.value)


def test_detect_nan_df():
    # checks that TypeError is raised when df contains object (str) data
    with pytest.raises(ValueError) as excinfo:
        uf.check_uniqueness(bad_unique_nan)
    # verifies TypeError contains appropriate message
    assert 'should not contain nulls' in str(excinfo.value)


def test_detect_nan_ser():
    # checks that TypeError is raised when df contains object (str) data
    with pytest.raises(ValueError) as excinfo:
        uf.check_uniqueness(bad_unique_nan['b1'])
    # verifies TypeError contains appropriate message
    assert 'should not contain nulls' in str(excinfo.value)


def test_check_uniqueness_good_df():
    # verifies that good data passes uniqueness check (bool)
    cols = ['dupes_present', 'dupe_count', 'prop_dupe']
    assert not uf.check_uniqueness(good).loc[:, cols].any(axis=None)


def test_check_uniqueness_good_ser():
    # verifies that good data passes uniqueness check (bool)
    cols = ['dupes_present', 'dupe_count', 'prop_dupe']
    assert not uf.check_uniqueness(good['g1']).loc[:, cols].any(axis=None)


def test_check_uniques_bad_df():
    # verifies that the uniqueness check find duplicates (bool)
    assert fail_dupe_df.equals(uf.check_uniqueness(bad_unique))


def test_check_uniques_bad_ser():
    # verifies that the uniqueness check find duplicates (bool)
    assert fail_dupe_ser.equals(uf.check_uniqueness(bad_unique['b1']))



# standard library imports
from typing import List, Optional, Union
# third party imports
import pandas as pd
# local imports
from datasurveyor import _utils


class GeneralFeatures:

    @staticmethod
    def check_nulls(data: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
        """Checks if data contains nulls.

        Args:
            data: Data to be checked for nulls.

        Returns:
            DataFrame with bool(s) indicating if data contains any nulls, count of the nulls
            present, and the proportion of nulls.
        """
        is_df = _utils.check_if_df(data)
        is_nulls = data.isna().any(axis=0)
        count_nulls = data.isna().sum(axis=0)
        if is_df:
            prop_nulls = count_nulls.divide(data.shape[0])
        else:
            prop_nulls = count_nulls / data.shape[0]
        result = _utils.result_to_df(
            data=is_nulls,
            title='nulls_present',
            null_count=count_nulls,
            prop_null=prop_nulls
        )
        return result

    @staticmethod
    def check_fuzzy_nulls(
            data: Union[pd.DataFrame, pd.Series],
            add_fuzzy_nulls: Optional[List] = None,
    ) -> pd.DataFrame:
        """Checks if DataFrame contains values commonly used to denote nulls (fuzzy nulls).

        Args:
            data: Data to be checked for fuzzy nulls.
            add_fuzzy_nulls: Additional items to check as fuzzy nulls.

        Returns:
            DataFrame with bool(s) indicating if data contains any fuzzy nulls, count of
            the fuzzy nulls present, and the proportion of fuzzy nulls.
        """
        is_df = _utils.check_if_df(data)
        fuzzy_nulls = ['null', 'Null', 'NULL', '', ' ']
        if add_fuzzy_nulls is not None:
            fuzzy_nulls.extend(add_fuzzy_nulls)
        is_fuzzy_nulls = data.isin(fuzzy_nulls).any(axis=0)
        count_fuzzy_nulls = data.isin(fuzzy_nulls).sum(axis=0)
        if is_df:
            prop_fuzzy_nulls = count_fuzzy_nulls.divide(data.shape[0])
        else:
            prop_fuzzy_nulls = count_fuzzy_nulls / data.shape[0]
        result = _utils.result_to_df(
            data=is_fuzzy_nulls,
            title='fuzzy_nulls_present',
            fuzzy_null_count=count_fuzzy_nulls,
            prop_fuzzy_null=prop_fuzzy_nulls,
        )
        return result



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['datasurveyor']

package_data = \
{'': ['*']}

install_requires = \
['pandas >=1.0.4']

extras_require = \
{'test': ['pytest >=5.4.3', 'pytest-cov >=2.10.0']}

setup(name='datasurveyor',
      version='0.0.1',
      description='Data exploration tools.',
      author='Nick Buker',
      author_email='nickbuker@gmail.com',
      url='https://github.com/nickbuker/datasurveyor',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='~=3.6,<4',
     )
